// (C) Copyright 2006 Douglas Gregor <doug.gregor -at- gmail.com>

// Use, modification and distribution is subject to the Boost Software
// License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

//  Authors: Douglas Gregor

/** @file environment.cpp
 *
 *  This file reflects the Boost.MPI "environment" class into Python
 *  methods at module level.
 */
#include <boostmpi/environment.hpp>
#include <boost/python.hpp>
#include <boost/weak_ptr.hpp>

using namespace boost::python;
using namespace boost::mpi;

namespace boost { namespace mpi { namespace python {

extern const char* environment_init_docstring;
extern const char* environment_finalize_docstring;
extern const char* environment_abort_docstring;
extern const char* environment_initialized_docstring;
extern const char* environment_finalized_docstring;
extern const char* env_docstring;




namespace
{
  struct argv_list
  {
    int m_argc;
    char** m_argv;

    argv_list()
      : m_argv(0)
    {
    }

    argv_list(boost::python::list py_argv)
    {
      m_argc = len(py_argv);
      m_argv = new char*[m_argc];
      for (int arg = 0; arg < m_argc; ++arg)
        m_argv[arg] = strdup(boost::python::extract<const char*>(py_argv[arg]));
    }

    void deallocate_argv()
    {
      if (m_argv)
      {
        for (int arg = 0; arg < m_argc; ++arg)
          free(m_argv[arg]);
        delete [] m_argv;
        m_argv = 0;
      }
    }

    ~argv_list()
    { deallocate_argv(); }
  };
}




static boost::weak_ptr<environment> env; 




boost::shared_ptr<environment> make_environment_1(bool abort_on_exception)
{
  if (environment::initialized())
    throw std::runtime_error("Boost.MPI already initialized");

  boost::shared_ptr<environment> result(new environment(abort_on_exception));

  env = result;
  return result;
}




boost::shared_ptr<environment> make_environment_2(
    boost::python::list argv,
    bool abort_on_exception)
{
  if (environment::initialized())
    throw std::runtime_error("Boost.MPI already initialized");

  argv_list alist(argv);
  int mpi_argc = alist.m_argc;
  char **mpi_argv = alist.m_argv;

  boost::shared_ptr<environment> result(new environment(
        mpi_argc, mpi_argv, abort_on_exception));

  // If anything changed, convert C-style argc/argv into Python argv
  if (mpi_argv != alist.m_argv)
    PySys_SetArgv(mpi_argc, mpi_argv);

  env = result;
  return result;
}




boost::shared_ptr<environment> current_environment()
{
  if (env.expired())
    return boost::shared_ptr<environment>();
  else
    return env.lock();
}




void export_environment()
{
  using namespace boost::python;
  using boost::python::arg;

  {
    typedef environment cl;
    class_<cl, boost::shared_ptr<cl> , boost::noncopyable>(
        "Environmnent", no_init)
      .def("__init__", make_constructor(make_environment_1,
            default_call_policies(),
            arg("abort_on_exception")=true),
            environment_init_docstring)
      .def("__init__", make_constructor(make_environment_2,
            default_call_policies(),
            (arg("argv"), arg("abort_on_exception")=true)),
            environment_init_docstring)

      .def("abort", &cl::abort, arg("errcode"), environment_abort_docstring)
      .staticmethod("abort")
      .def("initialized", &cl::abort, environment_initialized_docstring)
      .staticmethod("initialized")
      .def("finalized", &cl::abort, environment_finalized_docstring)
      .staticmethod("finalized")

      .def("_max_tag", &cl::max_tag)
      .staticmethod("_max_tag")
      .def("_collectives_tag", &cl::collectives_tag)
      .staticmethod("_collectives_tag")
      .def("_processor_name", &cl::processor_name)
      .staticmethod("_processor_name")

      .def("_host_rank", &cl::host_rank)
      .staticmethod("_host_rank")
      .def("_io_rank", &cl::io_rank)
      .staticmethod("_io_rank")
      ;
  }

  def("env", current_environment, env_docstring);
}

} } } // end namespace boost::mpi::python
